﻿SpeakinSpell = LibStub("AceAddon-3.0"):NewAddon("SpeakinSpell", "AceConsole-3.0", "AceEvent-3.0")

-------------------------------------------------------------------------------
-- CONSTANTS
-------------------------------------------------------------------------------


-- load localized text including default speeches
local L					= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell",					false)
local DEFAULT_SPEECHES	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)
local HELPFILE			= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_HELPFILE",			false)


SpeakinSpell.CURRENT_VERSION = "3.1.2.14"

local MAX_MESSAGES_PER_SPELL = 100 --do not ever make this number smaller, but we can always make it bigger
local URL = "http://www.wowace.com/projects/speakinspell/"

local FREQUENCY_DEFAULT = 1.00 -- means 100%
local COOLDOWN_DEFAULT = 0 -- default of 0 seconds = no cooldown


-- create a reverse lookup table to un-localize channel names
-- so we can return a channel name that is compatible with SendChatMessage()
local CHANNEL_TABLE = {
	[L["SAY"]]			= "SAY",
	[L["PARTY"]]		= "PARTY",
	[L["RAID"]]			= "RAID",
	[L["EMOTE"]]		= "EMOTE",
	[L["BATTLEGROUND"]]	= "BATTLEGROUND",
	[L["YELL"]]			= "YELL",
	[L["RAID_WARNING"]]	= "RAID_WARNING",
}


-------------------------------------------------------------------------------
-- CHANNEL RULES
-------------------------------------------------------------------------------


--NOTE: these need to be UN-localized chat channel names for use in SendChatMessage
local DEFAULT_CHANNEL_RULES = {
	BG		= nil,		--L["BATTLEGROUND"],
	Raid	= nil,		--L["RAID"],
	Party	= "SAY",	--L["PARTY"]
	Solo	= "SAY",
}


-------------------------------------------------------------------------------
-- DEBUG MESSAGES
-------------------------------------------------------------------------------


function SpeakinSpell:Print(message)
	-- Overriding AceConsole-3.0:Print to use localized addon name instead of tostring(self)
	local text = "|cff33ff99"..--[[tostring( self )]] L["SpeakinSpell"].."|r: " .. message
	DEFAULT_CHAT_FRAME:AddMessage( text )
end


function SpeakinSpell:DebugMsg(funcname, message)
	if SpeakinSpellSavedData and SpeakinSpellSavedData.ShowDebugMessages then
		if funcname then
			self:Print(funcname.."() - "..message)
		else
			self:Print(message)
		end
	end
end


function SpeakinSpell:ErrorMsg(funcname, message)
	self:Print("ERROR: "..funcname.."() - "..message)
end


function SpeakinSpell:DebugMsgDumpString(name, value)
	self:DebugMsg( nil, string.format("%s=%s", tostring(name), tostring(value)) )
end


function SpeakinSpell:DebugMsgDumpBool(name, value)
	if value then
		self:DebugMsg(nil, name.."=true")
	else
		self:DebugMsg(nil, name.."=false")
	end
end


function SpeakinSpell:DebugMsgDumpTable(Table, name)
	for k,v in pairs(Table) do
		if type(v) == "table" then
			self:DebugMsgDumpTable(v, name.."."..tostring(k))
		else
			self:DebugMsg( nil, name.."["..k.."]".."="..tostring(v) )
		end
	end
end


-------------------------------------------------------------------------------
-- UTILITY FUNCTIONS
-------------------------------------------------------------------------------


-- create a deep copy of the table and return it
function SpeakinSpell:CopyTable(SourceTable)
	local dest = {} -- make a new table in 'dest'
	for k,v in pairs(SourceTable) do
		-- dig into embedded tables to copy those too, not just create references
		if type(v) == "table" then
			dest[k] = self:CopyTable(v)
		else
			dest[k] = v
		end
	end
	return dest
end



function SpeakinSpell:GetDefaultTarget()
	-- try the currently selected target
	local target = UnitName("target")
	if target then
		self:DebugMsg(nil,"<target> is unknown, using selected:"..target)
		return target
	end
	
	-- nobody is selected, try focus
	target = UnitName("focus")
	if target then
		self:DebugMsg(nil,"<target> is unknown, using focus:"..target)
		return target
	end
	
	-- assume self-cast ability
	target = UnitName("player")
	self:DebugMsg(nil,"<target> is unknown, using self:"..target)
	
	return target
end


function SpeakinSpell:DeleteSpell(key)
	-- delete the current selection from the SavedData and RuntimeData tables
	SpeakinSpellSavedData.EventTable[key] = nil
	
	if not self.RuntimeData then
		-- we're still initializing, so the rest of this stuff doesn't matter
		return
	end
	
	-- remove runtime data for messaging
	self.RuntimeData.LastMessage[key] = nil
	
	-- update GUI controls
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[key] = nil
	
	-- update selected item
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
	
	-- restore the spell to the list of new detected
	-- so the user doesn't have to retrigger it
	-- self:RecordNewSpell(EventTableEntry.DetectedEvent)
	-- NO! could be outdated and bugged event
	-- just make the user recreate it	
	
	self:CurrentMessagesGUI_ValidateSelectedEvent() -- make sure we still have a valid spell selection
	self:CurrentMessagesGUI_Refresh()
end


function SpeakinSpell:ShowOptions()
	-- this is probably not necessary, but make sure this is up to date before opening the options panel
	self:CurrentMessagesGUI_Refresh()
	
	-- open interface options to show SpeakinSpell
	InterfaceOptionsFrame_OpenToCategory(L["SpeakinSpell"])
end


function SpeakinSpell:ShowHelp()
	-- this is probably not necessary, but make sure this is up to date before opening the options panel
	self:CurrentMessagesGUI_Refresh()
	
	-- open interface options to show SpeakinSpell
	InterfaceOptionsFrame_OpenToCategory(L["SpeakinSpell Help"])
end


function SpeakinSpell:ShowCreateNew()
	-- this is probably not necessary, but make sure this is up to date before opening the options panel
	self:CurrentMessagesGUI_Refresh()
	
	-- open interface options to show SpeakinSpell
	InterfaceOptionsFrame_OpenToCategory(L["Create New..."])
end


function SpeakinSpell:ShowMessageOptions()
	-- this is probably not necessary, but make sure this is up to date before opening the options panel
	self:CurrentMessagesGUI_Refresh()
	
	-- open interface options to show SpeakinSpell
	InterfaceOptionsFrame_OpenToCategory(L["Message Settings TITLE"])
end


function SpeakinSpell:Advertise(Channel, msgtarget)
	-- advertise into a channel for the kind of group you are in
	-- unless a channel has been specified
	if Channel == "" or Channel == nil then
		if (IsActiveBattlefieldArena()) then
			Channel = "PARTY" -- use Party chat for advertisements in Arena
		elseif (GetNumBattlefieldScores() > 0) then
			Channel = "SAY" -- don't advertise in BG chat, that would be annoying
		elseif (GetNumRaidMembers() > 0) then
			Channel = "RAID"
		elseif (GetNumPartyMembers() > 0) then
			Channel = "PARTY"
		else -- solo
			Channel = "SAY"
		end
	end
	
	-- if target is literally <target> or unspecified, then get the current target
	-- /ss advertise /w <target>
	-- /ss advertise /w
	if not msgtarget or msgtarget == "" or msgtarget == "<target>" or msgtarget == "<"..L["<target>"]..">" then
		msgtarget = self:GetDefaultTarget()
	end
	
	-- select a random ad that was not used last time
	local sizelist = #(DEFAULT_SPEECHES.ADVERTISEMENTS)
	local tempnum = math.random(1, sizelist);
	
	msg = DEFAULT_SPEECHES.ADVERTISEMENTS[tempnum]
	assert(msg)
	assert(type(msg) == "string")
	
	-- avoid repeating the same message we used last time
	-- unless there's only one in the list
	while sizelist > 1 and msg == self.RuntimeData.LastAd do
		tempnum = math.random(1, sizelist);
		msg = DEFAULT_SPEECHES.ADVERTISEMENTS[tempnum]
	end
	
	-- remember we selected this, so we don't repeat it next time
	self.RuntimeData.LastAd = msg

	-- process substitution variables on the randomized advertisements
	--TODOFUTURE: ?? change ads to a customizable macro event, to share more code with OnEventDetected
	local DetectedEventStub = {
		name = "ad",
		rank = "Silly",
		target = msgtarget,
		caster = UnitName("player"),
		type = "MACRO"
	}
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	msg = self:ProcessSubstitutionVars(msg, DetectedEvent)
	
	-- speak in chat
	SendChatMessage(msg, Channel, nil, msgtarget)
	SendChatMessage(URL, Channel, nil, msgtarget)	
end


function SpeakinSpell:TitleCase(s)
	local first = string.sub(s,1,1)
	local rest = string.sub(s,2,-1)

	first = string.upper(first)
	
	rest = string.gsub(rest, " %a",
		function(match) 
			return string.upper(match)
		end
	)
	
	local result = first..rest
	return result
end


function SpeakinSpell:StartsWith(s, prefix)
	start = string.sub(s, 1, string.len(prefix))
	return (start == prefix)
end


function SpeakinSpell:EndsWith(s, suffix)
	strend = string.sub(s, -1 * string.len(suffix))
	return (strend == suffix)
end


-------------------------------------------------------------------------------
-- DATA STRUCTURES (? not sure what to call this category)
-------------------------------------------------------------------------------


-- RecordNewEvent creates and maintains a list of new spells that don't have messages associated with them
-- this list is used to present the user with a list of options when she wants to start announcing a new spell
function SpeakinSpell:RecordNewEvent(DetectedEvent)
	local funcname = "RecordNewEvent"
	
	-- if we already recorded this spell, don't add a duplicate, unless this is a different rank
	-- NOTE: RecordNewEvent is called for both ranked, and rankless versions of the event
	if not self.RuntimeData.NewEventsDetected[DetectedEvent.key] then
		self:DebugMsg(funcname, DetectedEvent.SpellLink)
		self.RuntimeData.NewEventsDetected[DetectedEvent.key] = DetectedEvent
		-- update the options GUI if it exists
		if self.OptionsGUI then
			self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[DetectedEvent.key] = DetectedEvent.DisplayName
			table.sort(self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values)
			self.OptionsGUIStates.CreateNew.SelectedEventKey = DetectedEvent.key
			-- if this is a new macro, open the Create New interface
			if DetectedEvent.type == "MACRO" then
				self:ShowCreateNew()
			end
		end
	end
end


function SpeakinSpell:FormatDisplayName(de)
	-- create a display name for this event
	if de.rank ~= "" then
		return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." ("..de.rank..")"
	else
		if de.type == "UNIT_SPELLCAST_SENT" then
			return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." (Any Rank)" --TODOLATER: localize
		else --other supported events (buffs, debuffs, macros, and system events) donever 't have ranks
			return L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name
		end
	end
end


function SpeakinSpell:SafeGetSpellLink(de)
	local SpellLink,_ = GetSpellLink(de.name, de.rank)
	if SpellLink then
		return SpellLink
	else
		return "["..de.name.."]"
	end
end



-- returns rankedkey, ranklesskey
-- if not a rankable event, then ranklesskey will be nil
-- TODOFUTURE: this could return an ordered list of potential keys?
-- maybe for linking UNIT_SPELLCAST_SENT to the matching SPELL_AURA_GAIN 
-- and/or for making speech sets that apply to many spells with the same speech settings
function SpeakinSpell:MakeEventTableKeys( de )
	-- create a unique key for this event
	local rankedkey		= string.gsub( string.upper(de.type..de.name..de.rank), " ", "_" ) -- ACTIVE 3.1.2.07 format
	local ranklesskey	= string.gsub( string.upper(de.type..de.name),			" ", "_" ) -- ACTIVE 3.1.2.07 format
	
	if (de.type == "UNIT_SPELLCAST_SENT") and (rankedkey ~= ranklesskey) then
		return rankedkey, ranklesskey
	else
		return rankedkey, nil
	end
end


function SpeakinSpell:CreateDetectedEvent( DetectedEventStub )
	local funcname = "CreateDetectedEvent"
	
	local de = self:CopyTable(DetectedEventStub)
	
	-- create a spell link, if possible
	-- the stub will set SpellLink = true if the event is a spell
	-- else will we make a fake one
	de.SpellLink = self:SafeGetSpellLink(de)
	
	-- validate the target - get a backup target from selection, focus, or assumed self-cast
	if not de.target or de.target == "" or de.target == "Unknown" then
		de.target = self:GetDefaultTarget()
	end
	
	-- fetch additional target information
	de.targetclass = UnitClass(de.target)
	de.targetrace = UnitRace(de.target)

	--setup <pet>
	de.pet = UnitName("pet")
	
	-- create a display name for this event
	de.DisplayName = self:FormatDisplayName(de)
	
	return de
end


function SpeakinSpell:CheckSpellFrequency( de )
	-- check the frequency setting against a random number to determine if we should speak or not
	local Chance = math.random(1,100)
	local Frequency = de.EventTableEntry.Frequency*100

	if Chance > Frequency then
		self:DebugMsg( "CheckSpellFrequency", string.format("random chance failed for '%s' rolled %d > %d", de.SpellLink, Chance, Frequency) )
		return false
	end
	
	return true
end


function SpeakinSpell:CheckSpellCooldown( de )
	local funcname = "CheckSpellCooldown"
	
	-- check for a cooldown limit on announcing this spell

	-- NOTE: if we have not spoken for this spell yet, then
	--		self.RuntimeData.LastMessageTime or self.RuntimeData.LastMessageTime[spellname] might not exist
	--		in that case, set elapsed = cooldown, so that it will seem like enough time has passed
	local now = GetTime()
	if self.RuntimeData.LastMessageTime == nil then
		self.RuntimeData.LastMessageTime = {}
	end
	local last = self.RuntimeData.LastMessageTime[de.key]
	if last == nil then
		last = 0
	end
	local elapsed = de.EventTableEntry.Cooldown
	if last then
		elapsed = now - last
	end
	-- Now finally actually check if the cooldown is in effect
	if elapsed < de.EventTableEntry.Cooldown then
		-- the cooldown is in effect, so be silent
		self:DebugMsg( funcname, string.format("messaging on cooldown: %s: now:%d, last:%d, elapsed:%d, CD:%d", de.SpellLink, now, last, elapsed, de.EventTableEntry.Cooldown) )
		return false
	end
	
	return true
end


function SpeakinSpell:AllowSpeakForSpell( de )
	local funcname = "AllowSpeakForSpell"
	
	if de.EventTableEntry.DisableAnnouncements then
		self:DebugMsg(funcname,"OnEventDetected - messages for " .. de.SpellLink .. " are disabled")
		return false
	end
	
	if not SpeakinSpellSavedData.EnableAllMessages then
		self:DebugMsg(funcname,"OnEventDetected - all SpeakinSpell messages are disabled")
		return false
	end
	
	if not self:CheckSpellFrequency( de ) then
		return false
	end
	
	if not self:CheckSpellCooldown( de ) then
		return false
	end
	
	return true
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - GENERAL SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:GeneralOptions_OnShowSpellEventsToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowSpellEvents
	else -- "SET"
		SpeakinSpellSavedData.ShowSpellEvents = value
	end
end


function SpeakinSpell:GeneralOptions_OnShowDebugMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowDebugMessages
	else -- "SET"
		SpeakinSpellSavedData.ShowDebugMessages = value
	end
end


function SpeakinSpell:GeneralOptions_OnEnableAllMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.EnableAllMessages
	else -- "SET"
		SpeakinSpellSavedData.EnableAllMessages = value
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - MESSAGE SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:CurrentMessagesGUI_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if eventkey then
		return SpeakinSpellSavedData.EventTable[eventkey]
	else
		return nil
	end
end


function SpeakinSpell:CurrentMessagesGUI_ValidateSelectedEvent()
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		-- current selection is valid, so keep it
		return
	end
	
	-- select the first thing in the list, if possible
	for key,_ in pairs(SpeakinSpellSavedData.EventTable) do
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = key
		return
	end
	
	-- our spell list is empty
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
end



function SpeakinSpell:CurrentMessagesGUI_OnClickDelete()
	local funcname = "CurrentMessagesGUI_OnClickDelete"

	-- get the selected EventTableEntry
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if not eventkey then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- delete the current selection from the SavedData table
	self:DeleteSpell(eventkey)
end


function SpeakinSpell:CurrentMessagesGUI_OnClickGotoCreate()
	InterfaceOptionsFrame_OpenToCategory(L["Create New..."])
end


function SpeakinSpell:CurrentMessagesGUI_OnEventSelect(getset, value)
	if "GET" == getset then
		-- create a side-effect to update the shown/hidden state of all options GUI controls
		self:CurrentMessagesGUI_Refresh()
		-- return the current spell selection
		return self.OptionsGUIStates.MessageSettings.SelectedEventKey
	else -- "SET"
		-- update the current selection
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnChannelSelect(getset, scenario, value)
	local funcname = "CurrentMessagesGUI_OnChannelSelect"
	
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end

	if "GET" == getset then
		--self:DebugMsg(funcname,"GET "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
		if EventTableEntry.Channels[scenario] then
			return L[EventTableEntry.Channels[scenario]]
		else
			return L["Silent"]
		end
	else --"SET"
		-- store the new channel selection
		if value == L["Silent"] then
			EventTableEntry.Channels[scenario] = nil
		else
			EventTableEntry.Channels[scenario] = CHANNEL_TABLE[value]
		end
		--self:DebugMsg(funcname,"set "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
	end
end



function SpeakinSpell:CurrentMessagesGUI_OnWhisperTarget(getset, value)
	-- make sure we have a valid spell selection
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg( "CurrentMessagesGUI_OnWhisperTarget", "no event selected" )
		return false
	end
	
	if "GET" == getset then
		return EventTableEntry.WhisperTarget
	else
		EventTableEntry.WhisperTarget = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnMessage(getset, index, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnMessage", "no event selected")
		return ""
	end
	
	if "GET" == getset then
		local msg = EventTableEntry.Messages[index]
		if not msg then
			msg = ""
		end
		return msg
	else -- "SET"
		-- make sure we have a message table for the currently selected spell
		if not EventTableEntry.Messages then
			EventTableEntry.Messages = {}
		end
		
		-- store the new value into the message table
		EventTableEntry.Messages[index] = value
		self:DebugMsg( nil,"Message " .. tostring(index) .. " for spell " .. EventTableEntry.DetectedEvent.DisplayName .. " set to: " .. tostring(EventTableEntry.Messages[index]) )
		
		-- remove holes from the message table
		-- if the user has just deleted one of the entries, move everything else up to fill in the gap
		if value == "" then
			local Messages = EventTableEntry.Messages -- create reference for shortcut notation
			for i=index+1,MAX_MESSAGES_PER_SPELL,1 do
				Messages[i-1] = Messages[i]
			end
			Messages[MAX_MESSAGES_PER_SPELL] = nil
			-- destroy empty strings from the table
			for i=MAX_MESSAGES_PER_SPELL-1,1,-1 do
				if Messages[i] == "" then
					Messages[i] = nil
				else
					break --the rest should be valid non-empty strings
				end
			end
			-- hide the superfluous input boxes
			self:CurrentMessagesGUI_Refresh()
		end
	end
end




function SpeakinSpell:CurrentMessagesGUI_OnSpellDisable(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnSpellDisable", "no event selected")
		return
	end

	if "GET" == getset then
		return EventTableEntry.DisableAnnouncements
	else -- "SET"
		EventTableEntry.DisableAnnouncements = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnFrequencySelect(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnFrequencySelect", "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Frequency
	else --"SET"
		EventTableEntry.Frequency = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnCooldownSlider(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnCooldownSlider", "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Cooldown
	else --"SET"
		EventTableEntry.Cooldown = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_Refresh()
	local funcname = "CurrentMessagesGUI_Refresh"
	
	-- safety check in case this function is called early
	if not self.OptionsGUI then
		self:DebugMsg(funcname,"called prematurely, abort")
		return
	end
	
	-- determine how many Spell Message N input boxes are shown
	-- typically we'll show at least 1 speech message input box if we have a spell selected
	-- we want to show 1 more than the number of speeches currently configured
	-- unless we don't have a spell selected or we're in EnterNewSpell mode
	local NumSpellMessageInputsShown = 0
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		NumSpellMessageInputsShown = 1 + #(EventTableEntry.Messages)
	end
	self:DebugMsg(funcname,"NumSpellMessageInputsShown:"..tostring(NumSpellMessageInputsShown))
	
	-- show SpellMessageN input box if N <= NumSpellMessageInputsShown (numbering starts at 1)
	for i=1,MAX_MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.CurrentMessagesGUI.args["SpellMessage"..i].guiHidden = (i > NumSpellMessageInputsShown)
	end
	
	-- hide all the options if a spell is not selected
	local HideControls = false
	if not EventTableEntry then
		HideControls = true
	end
	
	self.OptionsGUI.args.CurrentMessagesGUI.args["DeleteSpellButton"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["Disable"				].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectCaption"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectBG"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectRaid"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectParty"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectSolo"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectArena"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["WhisperTargetToggle"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["FrequencySelect"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["CooldownSlider"		].guiHidden = HideControls
end



function SpeakinSpell:CurrentMessagesGUI_RebuildSpellList()
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values = {}
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
	for key,object in pairs(SpeakinSpellSavedData.EventTable) do
		-- build the spell list for use by the options GUI
		self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[key] = object.DetectedEvent.DisplayName
		-- make sure we have a current spell selected for the options GUI
		if not self.OptionsGUIStates.MessageSettings.SelectedEventKey then
			self.OptionsGUIStates.MessageSettings.SelectedEventKey = key
		end
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - CREATE NEW
-------------------------------------------------------------------------------


function SpeakinSpell:CreateNew_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.CreateNew.SelectedEventKey
	if eventkey then
		return self.RuntimeData.NewEventsDetected[eventkey]
	else
		return nil
	end
end


function SpeakinSpell:CreateNew_ValidateSelectedEvent()
	local DetectedEvent = self:CreateNew_GetSelectedEventObject()

	-- allow the current selection to remain if it is valid
	if DetectedEvent then
		return
	end

	-- select the first thing in the list, if possible
	for key,_ in pairs(self.RuntimeData.NewEventsDetected) do
		self.OptionsGUIStates.CreateNew.SelectedEventKey = key
		return
	end
	
	-- our spell list is empty
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
end


function SpeakinSpell:CreateNew_OnNewSpellNameSelect(getset, value)
	if "GET" == getset then
		--if nothing is selected, try to select the first thing in the list
		self:CreateNew_ValidateSelectedEvent()
		return self.OptionsGUIStates.CreateNew.SelectedEventKey
	else -- "SET"
		self.OptionsGUIStates.CreateNew.SelectedEventKey = value
	end
end


function SpeakinSpell:CreateNew_RebuildSpellList()
	self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values = {}
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
	for key,object in pairs(self.RuntimeData.NewEventsDetected) do
		self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[key] = object.DisplayName
		-- make sure we have a current spell selected for the options GUI
		if not self.OptionsGUIStates.CreateNew.SelectedEventKey then
			self.OptionsGUIStates.CreateNew.SelectedEventKey = object.key
		end
	end
end

function SpeakinSpell:CreateNew_OnClickCreateNew()
	local funcname = "CreateNew_OnClickCreateNew"
	
	-- grab the selected new event
	local de = self:CreateNew_GetSelectedEventObject()
	if not de then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- add the spell to the message table and related data structures
	-- enter the new spell into the spell tabe
	if not SpeakinSpellSavedData.EventTable[de.key] then
		SpeakinSpellSavedData.EventTable[de.key] = {
			-- event information
			DetectedEvent = {
				key			= de.key,
				name		= de.name,
				rank		= de.rank,
				DisplayName = de.DisplayName,
				SpellLink	= de.SpellLink,
				-- caster	= de.caster -- not stored in event table
				-- target	= de.target -- not stored in event table
				-- pet		= de.pet    -- not stored in event table
			},
			-- default options
			Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
			Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
			DisableAnnouncements = false,
			WhisperTarget = false,
			Frequency = FREQUENCY_DEFAULT,
			Cooldown = COOLDOWN_DEFAULT,
		}
	end
	
	-- destroy the new event detection since this is now a registered event
	self.RuntimeData.NewEventsDetected[de.key] = nil
	
	-- deselect it from the Create New screen and remove it from that UI
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
	self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[de.key] = nil

	-- enter the new spell name into the list for the Message Settings GUI, and select it
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[de.key] = de.DisplayName
	table.sort(self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values)
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = de.key
	
	-- go to the new event's settings in the options GUI
	InterfaceOptionsFrame_OpenToCategory(L["Message Settings TITLE"])
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - HELP PAGES
-------------------------------------------------------------------------------


function SpeakinSpell:HelpPages_Build()
	-- build help pages
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		local NextPageDisplayName = ""
		if i < #(HELPFILE.PAGEORDER) then
			NextPageDisplayName = HELPFILE.PAGEORDER[i+1]
		else --wrap to first page
			NextPageDisplayName = HELPFILE.PAGEORDER[1]
		end
		-- setup OptionGUI object
		self.OptionsGUI.args[FrameObjectName] = {
			order = 3,
			type = "group",
			name = "unused_HelpFrame_name",
			desc = "unused_HelpFrame_desc",
			args = {
				CaptionNormalMode = {
					order = 1,
					type = "description",
					name = HELPFILE.PAGES[ HELPFILE.PAGEORDER[i] ], --help CONTENTS
				},
				NextPage = {
					order = 2,
					type = "execute",
					width = "full",
					name = NextPageDisplayName,
					desc = L["Go to the next help topic"],
					func = function() InterfaceOptionsFrame_OpenToCategory(NextPageDisplayName) end
				}
			}, --end args
		} --end title
		
		-- add a button to the TOC page, linking to the child page we just created above
		self.OptionsGUI.args.HelpRoot.args[FrameObjectName] = {
			order = i+1,
			type = "execute",
			width = "full",
			name = FrameDisplayName,
			desc = FrameDisplayName, --L["Go to the first help topic"], --TODOLATER: tooltips different from labels, and cleanup L[] entry
			func = function() InterfaceOptionsFrame_OpenToCategory(FrameDisplayName) end
		}

	end
end


function SpeakinSpell:HelpPages_AddToBlizOptions()
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		--TODOFUTURE: push down to tier 3 -- tier 3 seems buggy, trying tier 2 instead
		--self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell Help"])
		self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell"])
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - SHARED UTILITIES
-------------------------------------------------------------------------------


function SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	-- AddToBlizOptions(appName="SpeakinSpell", display name, parent, ObjectName)
	-- NOTE: the FrameObjectName here represents self.OptionsGUI.args[FrameObjectName]
	self.optionsFrames[FrameObjectName] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpeakinSpell", FrameDisplayName, ParentDisplayName, FrameObjectName)
	
	-- for some reason the framework has a problem if we define the "reset defaults" button handler in the data before calling AddToBlizOptions
	-- so we have to add it afterwards
	self.optionsFrames[FrameObjectName].default = function() self:ResetToDefaults(FrameObjectName) end
end


function SpeakinSpell:InitOptionsGUIState() -- sets up event lists and current selections
	self.OptionsGUIStates = {
		MessageSettings = {
			SelectedEventKey = nil,
		},
		CreateNew = {
			SelectedEventKey = nil,
		}
	}

	-- the options GUI may still have a reference to the old SavedData table.
	-- refresh it since we've rebuilt the table
	-- but if we're initializing then the OptionsGUI may not have been created yet
	if self.OptionsGUI then
		self:CurrentMessagesGUI_RebuildSpellList()
		self:CreateNew_RebuildSpellList()
	end
end


function SpeakinSpell:SetupOptionsGUI_CreateMessageInputs()
	-- add MAX_MESSAGES_PER_SPELL random speech input boxes
	for i=1,MAX_MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.CurrentMessagesGUI.args["SpellMessage"..i] = {
			order = i + 17, -- should be the order of the last item listed above the message input boxes (CooldownSlider), plus i
			width = "full",
			type = "input",
			name = string.format(L["Random Speech i"], i), -- "Random Speech %d"
			desc = L["Random Speech i TOOLTIP"],
			get = function() return  self:CurrentMessagesGUI_OnMessage("GET",i,nil) end,
			set = function(_, value) self:CurrentMessagesGUI_OnMessage("SET",i, value) end,
			guiHidden = true, --by default, until CurrentMessagesGUI_Refresh is called
		}
	end
end



function SpeakinSpell:RefreshFrame(FrameName)
	-- the only way I found to do this is to hide and then show the frame
	-- this occurs fast enough that I do not see a flicker when this happens
	-- UPDATE: saw a forum post when I was at work with the proper API call, but I already forget what it was, and this still works
	if self.optionsFrames and self.optionsFrames[FrameName] and self.optionsFrames[FrameName]:IsVisible() then
		self.optionsFrames[FrameName]:Hide()
		self.optionsFrames[FrameName]:Show()
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - ENTRY POINT FOR ALL
-------------------------------------------------------------------------------


function SpeakinSpell:SetupOptionsGUI()
	-- Define Options table for the GUI
	self.OptionsGUI = {
		type = "group",
		name = L["SpeakinSpell"],
		args = {
	---------------------------------------------------------------------------------------
			General = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					Caption = {
						order = 1,
						type = "description",
						name = L["General Settings"],
					},
					ShowURL = {
						order = 2,
						type = "description",
						name = URL,
					},
					EnableAllMessagesToggle = {
						order = 3,
						type = "toggle",
						width = "full",
						name = L["Enable All Messages"],
						desc = L["Enable All Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnEnableAllMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnEnableAllMessagesToggle("SET",value) end,
					},
					ShowSpellEventsToggle = {
						order = 4,
						type = "toggle",
						width = "full",
						name = L["Report detected spell events"],
						desc = L["Report detected spell events TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowSpellEventsToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowSpellEventsToggle("SET",value) end,
					},
					ShowDebugMessagesToggle = {
						order = 5,
						type = "toggle",
						width = "full",
						name = L["Show Debugging Messages (verbose)"],
						desc = L["Show Debugging Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowDebugMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowDebugMessagesToggle("SET",value) end,
					},
				}, --end args
			}, --end General

	---------------------------------------------------------------------------------------
			CurrentMessagesGUI = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = L["Message Settings"],
					},
					EventSelect = {
						order = 2,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP"],
						values = {},
						get = function() return  self:CurrentMessagesGUI_OnEventSelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnEventSelect("SET",value) end,
					},
					DeleteSpellButton = {
						order = 3,
						type = "execute",
						name = L["Delete Spell"],
						desc = L["Delete Spell TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickDelete() end,
					},
					CreateSpellButton = {
						order = 3,
						type = "execute",
						name = L["Create New..."],
						desc = L["Create New... TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickGotoCreate() end,
					},
					Disable = {
						order = 5,
						type = "toggle",
						width = "full",
						name = L["Disable announcements for this spell"],
						desc = L["Disable announcements for this spell TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnSpellDisable("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnSpellDisable("SET",value) end,
					},
					ChannelSelectCaption = {
						order = 6,
						type = "description",
						width= "full",
						name = L["Select the channels to use when announcing this spell when"],
					},
					ChannelSelectBG = {
						order = 10,
						type = "select",
						name = L["In a Battleground"],
						desc = L["In a Battleground TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","BG",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","BG",value) end,
					},
					ChannelSelectRaid = {
						order = 11,
						type = "select",
						name = L["In a Raid"],
						desc = L["In a Raid TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Raid",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Raid",value) end,
					},
					ChannelSelectParty = {
						order = 12,
						type = "select",
						name = L["In a Party"],
						desc = L["In a Party TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Party",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Party",value) end,
					},
					ChannelSelectSolo = {
						order = 13,
						type = "select",
						name = L["By yourself"],
						desc = L["By yourself TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Solo",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Solo",value) end,
					},
					ChannelSelectArena = {
						order = 14,
						type = "select",
						name = L["In Arena"],
						desc = L["In Arena TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Arena",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Arena",value) end,
					},
					WhisperTargetToggle = {
						order = 15,
						type = "toggle",
						width = "full",
						name = L["Whisper the message to your target"],
						desc = L["Whisper the message to your target TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnWhisperTarget("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnWhisperTarget("SET",value) end,
					},
					FrequencySelect = {
						order = 16,
						type = "range",
						width = "full",
						name = L["How often do you want to announce this spell?"],
						desc = L["How often TOOLTIP"],
						min = 0.01, max = 1.00, step=0.01,
						isPercent = true,
						get = function() return  self:CurrentMessagesGUI_OnFrequencySelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnFrequencySelect("SET",value) end,
					},
					CooldownSlider = {
						order = 17,
						type = "range",
						width = "full",
						name = L["Cooldown on announcements for this spell"] ,
						desc = L["Cooldown TOOLTIP"],
						min =0, max = 600, step=1, -- allowing for 0 seconds up to 10 minutes
						get = function() return  self:CurrentMessagesGUI_OnCooldownSlider("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnCooldownSlider("SET",value) end,
					},
				}, --end args
			}, --end CurrentMessagesGUI
			
	---------------------------------------------------------------------------------------
			Trade = {
				order = 1,
				type = "group",
				name = "unused_Trade_name",
				desc = "unused_Trade_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = "Coming Soon... randomized trade spam for all of your professional and class services, portals, LFG, LFW",
					},
				}, --end args
			}, --end Trade
			
	---------------------------------------------------------------------------------------
			CreateNew = {
				order = 1,
				type = "group",
				name = "unused_CreateNew_name",
				desc = "unused_CreateNew_desc",
				args = {
					NewSpellModeInstruct = {
						order = 1,
						type = "description",
						name = L["NEW SPELL INSTRUCTIONS"],
					},
					NewSpellNameSelect = {
						order = 2,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP - Create New"],
						values = {},
						get = function() return  self:CreateNew_OnNewSpellNameSelect("GET",nil) end,
						set = function(_, value) self:CreateNew_OnNewSpellNameSelect("SET",value) end,
					},
					CreateNewSpellButton = {
						order = 3,
						type = "execute",
						width = "full",
						name = L["Create Selected Event"],
						desc = L["Create Selected Event TOOLTIP"],
						func = function() self:CreateNew_OnClickCreateNew() end,
					},
				}, --end args
			}, --end CreateNew

	---------------------------------------------------------------------------------------
			HelpRoot = {
				order = 2,
				type = "group",
				name = "unused_HelpRoot_name",
				desc = "unused_HelpRoot_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = HELPFILE.ROOT,
					},
				}, --end args
			}, --end HelpRoot
			
		}, --end args
	} -- end OptionsGUI
	---------------------------------------------------------------------------------------
	
	-- set up spell/event lists and current selections
	self:InitOptionsGUIState()
	
	-- add MAX_MESSAGES_PER_SPELL random speech input boxes
	self:SetupOptionsGUI_CreateMessageInputs()
	
	-- build help pages
	self:HelpPages_Build()
	
	-- add the options to the in-game interface options panel at Escape > Interface > Addons > SpeakinSpell
	-- The ordering here matters, it determines the order in the Blizzard Interface Options
	-- (or it would matter if we had sub-categories, which we currently don't)
	-- InterfaceOptions_AddCategory(self.OptionsGUI) -- done via Ace below

	-- :RegisterOptionsTable(appName, options, slashcmd, persist)
	-- AceConfig:RegisterOptionsTable("MyAddon", myOptions, {"/myslash", "/my"})
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpeakinSpell", self.OptionsGUI, nil)
	
	--SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	--frame objects are stored to self.optionsFrames[FrameObjectName]
	self.optionsFrames = {}
	self:AddToBlizOptions("General",		L["SpeakinSpell"],		nil)
	self:AddToBlizOptions("CurrentMessagesGUI",	L["Message Settings TITLE"],		L["SpeakinSpell"])
	self:AddToBlizOptions("CreateNew",		L["Create New..."],		L["SpeakinSpell"])
	self:AddToBlizOptions("Trade",			L["Trade"],				L["SpeakinSpell"])
	self:AddToBlizOptions("HelpRoot",		L["SpeakinSpell Help"],	L["SpeakinSpell"])
	
	-- add help pages
	self:HelpPages_AddToBlizOptions()

end



-------------------------------------------------------------------------------
-- GENERAL INITIALIZATION
-------------------------------------------------------------------------------


-- our saved data is stored in SpeakinSpellSavedData.EventTable
-- we must declare the parent table SpeakinSpellSavedData
-- but we know whether real data has been loaded or not if SpeakinSpellSavedData.EventTable exists
-- NOTE: SpeakinSpellSavedData.EventTable may be an empty table
SpeakinSpellSavedData = {}

function SpeakinSpell:InitDefaultSavedData()
	-- copy defaults into current settings using class-specific defaults
	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
	SpeakinSpellSavedData.ShowSpellEvents = false
	SpeakinSpellSavedData.ShowDebugMessages = false
	SpeakinSpellSavedData.EnableAllMessages = true
	
	SpeakinSpellSavedData.EventTable = nil
	SpeakinSpellSavedData.EventTable = {}
	SpeakinSpellSavedData.EventTable = self:CopyTable( DEFAULT_SPEECHES.CLASSES[englishClass] )
	
	-- remove faction-specific spells and reset runtime data such as localizable display names
	for key, EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		-- rebuild some info that's normally built dynamically
		local de = EventTableEntry.DetectedEvent
		de.DisplayName = self:FormatDisplayName(de)
		de.key = key
		de.SpellLink = self:SafeGetSpellLink(de)
		
		if EventTableEntry.Faction ~= nil then --TODONOW: faction stuff has been lost - do something about it
			if EventTableEntry.Faction ~= UnitFactionGroup("player") then
				SpeakinSpellSavedData.EventTable[key] = nil
			end
		end
	end
end


function SpeakinSpell:ResetToDefaults(FrameName) --TODOFUTURE: reset only the current frame's defaults
	SpeakinSpellSavedData.EventTable = nil
	
	-- copy defaults into current settings using class-specific defaults
	self:InitDefaultSavedData()
	
	-- update/reset runtime data state
	self:InitRuntimeData()
	self:CurrentMessagesGUI_Refresh() -- this might be redundant but doesn't seem to hurt
	
	-- if the options GUI is showing right now, we need to refresh it to make it show the new data
	self:RefreshFrame("General")
	self:RefreshFrame("CreateNew")
	self:RefreshFrame("CurrentMessagesGUI")
	
	self:Print(L["Default settings restored"])
end


function SpeakinSpell:InitRuntimeData()
	-- this is just a declaration
	self.RuntimeData = {
		NewEventsDetected = {},		-- a list of new spells that the user may add
		LastMessage = {},			-- just an empty table for now, this will be [spellname] = "last random speech triggered by this spell" to prevent repeats
	}

	self:InitOptionsGUIState() -- sets up event lists and current selections
end


function SpeakinSpell:OnInitialize()
    -- Called when the addon is loaded

	self:RegisterAllEvents()
	self:RegisterChatCommands()	-- register chat commands
	-- the rest of the initialization does not occur until VARIABLES_LOADED
	
	-- done loading
    self:Print( string.format(L["Loaded version %s"], self.CURRENT_VERSION) )
end


function SpeakinSpell:RegisterAllEvents()
	-- register for spellcasting events
	-- which is our hook too know when to speak for a spell
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("VARIABLES_LOADED")
end


function SpeakinSpell:OnEnable()
	-- Called when the addon is enabled
	-- self:RegisterAllEvents() -- did this OnInitialize
	self:CurrentMessagesGUI_Refresh() -- make sure the options GUI is setup correctly
end


function SpeakinSpell:OnDisable()
    -- Called when the addon is disabled
	-- we don't need to do anything here
end


-------------------------------------------------------------------------------
-- UPDATE OLD SAVED DATA
-------------------------------------------------------------------------------


--[[
SpeakinSpell:ConvertOldSpellTableToNewEventTable()

-----------------------------------------------------
-- OLD FORMAT: 3.1.2.03
local spellnamekey = name
if rank then
	spellnamekey = name.."("..rank..")"
end
-- rank can also be "_ event: Login" or "_ macro something"

SpeakinSpellSavedData.EventTable[spellnamekey] = {
	-- default options
	Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
	WhisperTarget = false,
	Frequency = FREQUENCY_DEFAULT,
	Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
}

-----------------------------------------------------
-- NEW FORMAT: 3.1.2.05
-- DetectedEvent table has been added
-- key has changed
-- new key format is name..rank..type

SpeakinSpellSavedData.EventTable[de.key] = {
	-- event information
	DetectedEvent = {
		key			= de.key,
		name		= de.name,
		rank		= de.rank,
		type		= de.type,
		DisplayName = de.DisplayName,
		SpellLink	= de.SpellLink,
		-- caster	= de.caster -- not stored in event table
		-- target	= de.target -- not stored in event table
		-- pet		= de.pet    -- not stored in event table
	},
	-- default options
	Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
	WhisperTarget = false,
	Frequency = FREQUENCY_DEFAULT,
	Cooldown = 0,
	Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
}
--]]

-- oldkey is a v3.1.2.03 SpellTable key
-- returns a new DetectedEventStub object for 3.1.2.05
function SpeakinSpell:FixOld_31205_CreateDEStub(oldkey)

	local stub = {
		name = "",
		rank = "",
		type = "", --unknown
		-- in 3.1.2.03, the only detected events were self cast on player
		-- or equivalent for as much as we cared at the time.
		caster = UnitName("player"),
		target = UnitName("player"),
	}
	
	-- check for macros using the 3.1.2.03 method of key names
	if self:StartsWith(oldkey, "_ macro") or self:StartsWith(oldkey, "_ "..L["/macro"]) then
		--remove leading "_ " from the name
		stub.name = string.sub(oldkey, 3)
		stub.rank = ""
		stub.type = "MACRO"
		return stub
	end
	
	-- check for the only event that was supported in 3.1.2.03
	if oldkey == "_ event: Login" then
		stub.name = L.EVENTNAMES["SpeakinSpell Loaded"]
		stub.rank = ""
		stub.type = "EVENT"
		return stub
	end
	
	-- it must be either a UNIT_SPELLCAST_SENT or a SPELL_AURA_APPLIED_BUFF
	-- let's take a closer look... 
	
	-- GetSpellInfo
	local name, rank, icon, cost, isFunnel, powerType, castTime, minRange, maxRange = GetSpellInfo(oldkey)

	-- check whether they old key specified the rank or not
	local oldkeyrank = string.match(oldkey, "%b()")
	
	-- if GetSpellInfo returns data, then it's a real spell, which will mean UNIT_SPELLCAST_SENT
	if name then
		stub.type = "UNIT_SPELLCAST_SENT"
		stub.name = name
		if oldkeyrank and rank then
			stub.rank = rank
		else -- use unspecified rank
			stub.rank = ""
		end
		return stub
	end

	-- else, do some manual parsing and see what we can come up with
	
	-- parse the old key name for the name and rank
	stub.name = oldkey
	stub.rank = string.match(oldkey, "%b()")
	
	if not stub.rank then
		stub.rank = ""
	end
	
	if stub.rank ~= "" then
		stub.name = string.sub(oldkey, 1, -2 - string.len(stub.rank) ) -- remove rank from name
		stub.rank = string.sub(stub.rank, 2, -2) --remove wrapping parens
	end
	-- else the name is the full key, already set up above
	
	stub.name = self:TitleCase(stub.name)
	stub.rank = self:TitleCase(stub.rank)
	
	-- buffs received have no rank
	-- if this has rank, it's definitely not a buff
	if stub.rank ~= "" then
		stub.type = "UNIT_SPELLCAST_SENT"
		return stub
	end
	
--	self:Print("---------")
--	self:Print("GetSpellInfo - assumed buff")
--	self:Print("name:"..tostring(name))
--	self:Print("rank:"..tostring(rank))
--	self:Print("icon:"..tostring(icon))
--	self:Print("cost:"..tostring(cost))
--	self:Print("isFunnel:"..tostring(isFunnel))
--	self:Print("powerType:"..tostring(powerType))
--	self:Print("castTime:"..tostring(castTime))
--	self:Print("minRange:"..tostring(minRange))
--	self:Print("maxRange:"..tostring(maxRange))
--	self:Print("---------")

	-- can't tell for sure, so use both
	stub.type = "BOTH"
	
	return stub
end



function SpeakinSpell:FixOld_31205_CreateDetectedEvent( DetectedEventStub )

	local de = self:CopyTable(DetectedEventStub)
	
	-- create a spell link, if possible
	-- the stub will set SpellLink = true if the event is a spell
	-- else will we make a fake one
	if de.SpellLink then
		de.SpellLink,_ = GetSpellLink(de.name, de.rank)
	end
	-- NOTE: some aura gains are not linkable, notably "Refreshment" from "Drink"
	--		in that case, GetSpellLink will return NULL
	if not de.SpellLink then
		de.SpellLink = "["..de.name.."]"
	end
	
	--[[ don't need this for importing saved data
	
	-- validate the target - get a backup target from selection, focus, or assumed self-cast
	if not de.target or de.target == "" or de.target == "Unknown" then
		de.target = self:GetDefaultTarget()
	end
	
	-- fetch additional target information
	de.targetclass = UnitClass(de.target)
	de.targetrace = UnitRace(de.target)

	--setup <pet>
	de.pet = UnitName("pet")
	--]]
	
	-- create a display name for this event
	if de.rank ~= "" then
		de.DisplayName = L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." ("..de.rank..")"
	else
		if de.type == "UNIT_SPELLCAST_SENT" then
			de.DisplayName = L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." (Any Rank)" --TODOLATER: localize
		else --other supported events (buffs, debuffs, macros, and system events) donever 't have ranks
			de.DisplayName = L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name
		end
	end
	
	-- create a unique key for this event
	de.key = de.name..de.rank..de.type --DO NOT CHANGE 3.1.2.05 format is intentional
	
	--[[ don't need this for importing saved data
	
	-- check for a EventTableEntry for this specific rank
	de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
	
	-- if that wasn't found, check for a rankless entry
	if not de.EventTableEntry then
		local ranklesskey = de.name..de.type
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		-- if that was found, then update the key for future reference
		if de.EventTableEntry then
			de.key = ranklesskey
		end
	end
	--]]
	
	return de
end


function SpeakinSpell:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
	SpeakinSpellSavedData.EventTable[de.key] = {
		-- event information
		DetectedEvent = {
			key			= de.key,
			name		= de.name,
			rank		= de.rank,
			type		= de.type,
			DisplayName = de.DisplayName,
			SpellLink	= de.SpellLink,
			-- caster	= de.caster -- not stored in event table 31205
			-- target	= de.target -- not stored in event table 31205
			-- pet		= de.pet    -- not stored in event table 31205
		},
		-- default options
		-- Channels and Messages tables are guaranteed to exist
		-- the rest might not - correct for that below
		Channels		= self:CopyTable(OldSpellTableEntry.Channels),
		Messages		= self:CopyTable(OldSpellTableEntry.Messages),
	}
	
	-- validate missing data and set defaults
	-- this used to be performed as a side effect in the GUI prior to 3.1.2.05
	local de = SpeakinSpellSavedData.EventTable[de.key]
	
	--WhisperTarget
	if OldSpellTableEntry.WhisperTarget then
		de.WhisperTarget = OldSpellTableEntry.WhisperTarget
	else
		de.WhisperTarget = false
	end
	
	--Frequency
	if OldSpellTableEntry.Frequency then
		de.Frequency = OldSpellTableEntry.Frequency
	else
		de.Frequency = FREQUENCY_DEFAULT
	end
	
	--Cooldown
	if OldSpellTableEntry.Cooldown then
		de.Cooldown	= OldSpellTableEntry.Cooldown
	else
		de.Cooldown	= COOLDOWN_DEFAULT
	end
	
	--DisableAnnouncements
	if OldSpellTableEntry.DisableAnnouncements then
		de.DisableAnnouncements	= OldSpellTableEntry.DisableAnnouncements
	else
		de.DisableAnnouncements	= false
	end

	--self:Print("Imported: "..oldkey.." -> "..de.DisplayName)
end


function SpeakinSpell:FixOld_31205_ImportSpellTable()

	local dupwarn = false

	SpeakinSpellSavedData.EventTable = {}

	for oldkey,OldSpellTableEntry in pairs(SpeakinSpellSavedData.SpellTable) do
	
		-- the process in 31205 is basically 3 steps
		-- 1) declare a DetectedObjectStub with name, rank, type, caster, and target
		--		we will create this by parsing the oldkey value, which is where this info lived
		-- 2) pass that stub through the function FullDetectedEvent = CreateDetectedEvent(stub)
		--		we have a copy of that function, with some small tweaks and simplifications
		-- 3) create an EventTableEntry object out of the FullDetectedEvent, which adds the Messages and other settings
		local stub = self:FixOld_31205_CreateDEStub(oldkey)
		
		-- some spells are clearly in the spellbook (UNIT_SPELLCAST_SENT) 
		-- while others must be items (UNIT_SPELLCAST_SENT)  or procs (SPELL_AURA_APPLIED_BUFF)
		if stub.type ~= "BOTH" then
			-- stub.type is known
			local de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
		else
			-- stub.type is unknown, record both
			dupwarn = true
			self:Print("Unsure how to update: "..oldkey)
			
			stub.type = "UNIT_SPELLCAST_SENT"
			local de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)

			stub.type = "SPELL_AURA_APPLIED_BUFF"
			de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
		end
	end
	
	if dupwarn then
		self:Print("WARNING: Some of your saved settings had to be duplicated under two different events (both UNIT_SPELLCAST_SENT and SPELL_AURA_APPLIED_BUFF). You may want to remove the undesired one of the 2.")
		self.ShowSettingsPageOnLoad = L["Message Settings TITLE"]
	end
	
	-- we can destroy the old SpellTable now.  We won't be using it anymore
	SpeakinSpellSavedData.SpellTable = nil
end


function SpeakinSpell:FixOld_31206_UpdateEventKeys()
	--changing SpeakinSpellSavedData.EventTable causes a problem while iterating on it
	--so build a new table and use it to replace the old table
	local newkeys = {}
	
	if not SpeakinSpellSavedData.EventTable then
		self:Print("ERROR: SpeakinSpellSavedData.EventTable is nil, aborting update to 3.1.2.06")
		return
	end
	
	for oldkey,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		local de = EventTableEntry.DetectedEvent
		-- DO NOT CHANGE 3.1.2.06 format is intentional for upgrade chain
		de.key = de.type..de.name..de.rank -- DO NOT CHANGE 3.1.2.06 format
		newkeys[de.key] = EventTableEntry
	end
	
	SpeakinSpellSavedData.EventTable = self:CopyTable(newkeys)
end


function SpeakinSpell:FixOld_31207_UpdateEventKeys()
	--changing SpeakinSpellSavedData.EventTable causes a problem while iterating on it
	--so build a new table and use it to replace the old table
	local newkeys = {}
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		local de = EventTableEntry.DetectedEvent
		de.key = string.gsub( string.upper(de.type..de.name..de.rank), " ", "_" ) -- DO NOT CHANGE 3.1.2.07 format
		newkeys[de.key] = EventTableEntry
	end
	SpeakinSpellSavedData.EventTable = self:CopyTable( newkeys )
end


function SpeakinSpell:FixOldSavedData()

	-----------------------------------------------------------------------------------
	-- prior to v3.0.3.08, a version number was not saved in the SpeakinSpellSavedData
	-- create an older version number
	if not SpeakinSpellSavedData.Version then
		SpeakinSpellSavedData.Version = "3.0.3.07"
	end
	
	-----------------------------------------------------------------------------------
	-- Define Patches by version
	
	local Patches = {
		
		{ Version = "3.1.1.02", 
			-- Added this global toggle in 3.1.1.02, before this, it behaved as if always on
			func = function() SpeakinSpellSavedData.EnableAllMessages = true end 
		},
		
		{ Version = "3.1.2.05", 
			-- redefined the SpellTable as EventTable with a new key system and data format
			func = function() self:FixOld_31205_ImportSpellTable() end
		},
		
		{ Version = "3.1.2.06", 
			-- changed event keys so they sort by event type instead of spell name
			func = function() self:FixOld_31206_UpdateEventKeys() end
		},
		
		{ Version = "3.1.2.07",
			-- changed event keys to uppercase, with underscores replacing spaces
			func = function() self:FixOld_31207_UpdateEventKeys() end
		},
	}
	
	-----------------------------------------------------------------------------------
	-- Apply all patches
	
	for i, Patch in ipairs(Patches) do
		if SpeakinSpellSavedData.Version < Patch.Version then
			self:Print( string.format("Updating saved data %s -> %s", SpeakinSpellSavedData.Version, Patch.Version) )
			Patch.func() -- apply the patch defined above
			SpeakinSpellSavedData.Version = Patch.Version -- patch has been applied
		end
	end
	
	-----------------------------------------------------------------------------------
	-- saved data is now up to date with the current version
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
end


-------------------------------------------------------------------------------
-- EVENT HANDLERS
-------------------------------------------------------------------------------



function SpeakinSpell:VARIABLES_LOADED()
	local funcname = "VARIABLES_LOADED"
	self:DebugMsg(funcname, "Variables Loaded")
	
	-- SpeakinSpellSavedData will always exist
	-- SpeakinSpellSavedData.EventTable only exists if actual data has been loaded
	-- NOTE: SpeakinSpellSavedData.EventTable may be empty, but its parent table SpeakinSpellSavedData never will be
	if	SpeakinSpellSavedData.EventTable or -- current name
		SpeakinSpellSavedData.SpellTable then --old name from v3.1.2.04 and earlier
		-- If the format of the saved data has changed due to upgrading to a newer version
		-- then fix the old saved data for compatibility
		self:FixOldSavedData()
	else
		self:DebugMsg(funcname, "EventTable does not exist, using defaults")
		-- new user: load defaults
		self:InitDefaultSavedData()
	end

	-- set up runtime data that we don't save, only after the saved variables are loaded
	self:InitRuntimeData()
	
	-- set up the Options GUI config - we need the Runtime Data before we can do this
	self:SetupOptionsGUI()
	
	-- trigger a Login event
	-- this can be used for random messages
	self:DebugMsg(funcname, "Signal System Event: SpeakinSpell Loaded")
	local DetectedEventStub = {
		name = L.EVENTNAMES["SpeakinSpell Loaded"],
		rank = "",
		type = "EVENT",
		-- treat these special events as cast by the player on himself
		caster = UnitName("player"),
		target = UnitName("player"),
	}
	self:OnEventDetected( DetectedEventStub ) 
	
	-- if FixOldSavedData wants the user to review the updated data
	-- show the requested options frame now
	if self.ShowSettingsPageOnLoad then
		InterfaceOptionsFrame_OpenToCategory(self.ShowSettingsPageOnLoad)
		self.ShowSettingsPageOnLoad = nil
	end
	
end


function SpeakinSpell:SPELL_AURA_APPLIED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	local funcname = "SPELL_AURA_APPLIED"
	self:DebugMsg(funcname, "( srcName, dstName, auraType, spellName )")
	self:DebugMsgDumpString("srcName",srcName)
	self:DebugMsgDumpString("dstName",dstName)
	self:DebugMsgDumpString("auraType",auraType)
	self:DebugMsgDumpString("spellName",spellName)
	
	-- we're only concerned with buffs we receive
	if dstName ~= UnitName("player") then
		self:DebugMsg(funcname, "not cast on you")
		return
	end
	
	local DetectedEventStub = {
		name = spellName,
		rank = "", -- no rank known
		-- treat these special events as cast by the player on himself
		caster = srcName,
		target = dstName,
		type = "SPELL_AURA_APPLIED_"..auraType,
		SpellLink = true,
	}
	
	-- buffs cast on me by someone other than me are treated differently than buffs I cast on myself
	if srcName ~= UnitName("player") then
		self:DebugMsg(funcname, "not cast by you")
		DetectedEventStub.type = DetectedEventStub.type.."_FOREIGN"
	end
	
	-- process the spell
	-- this function is shared with the UNIT_SPELLCAST_SENT event handler
	self:OnEventDetected( DetectedEventStub )
end


function SpeakinSpell:COMBAT_LOG_EVENT_UNFILTERED(_,timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	--self:DebugMsg("COMBAT_LOG_EVENT_UNFILTERED",tostring(eventtype))
	if eventtype == "SPELL_AURA_APPLIED" then
		self:SPELL_AURA_APPLIED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	end
end


function SpeakinSpell:UNIT_SPELLCAST_SENT(event, caster, name, rank, target)
	local funcname = "UNIT_SPELLCAST_SENT"

	self:DebugMsg(funcname,"(caster, name, rank, target)")
	--self:DebugMsgDumpString("event",event) -- just echoes UNIT_SPELLCAST_SENT every time
	self:DebugMsgDumpString("caster",caster)
	self:DebugMsgDumpString("name",name)
	self:DebugMsgDumpString("rank",rank)
	self:DebugMsgDumpString("target",target)
	
	-- only speak for spells cast by the player, not anyone else
	if caster == "player" then
		caster = UnitName("player")
	else
		self:DebugMsg(funcname,"caster is not player")
		return
	end
	
	-- process the spell
	local DetectedEventStub = {
		name = name,
		rank = rank,
		caster = caster,
		target = target,
		type = "UNIT_SPELLCAST_SENT",
		SpellLink = true,
	}
	self:OnEventDetected( DetectedEventStub )
end


function SpeakinSpell:OnEventDetected( DetectedEventStub )
	local funcname = "OnEventDetected"

	-- fill in the stub
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	
	local rankedkey, ranklesskey = self:MakeEventTableKeys( DetectedEvent )
	
	local SpeakForRankless = true
	
	--NOTE: it looks like I want to make a subroutine here and call it twice
	--		but the logic is actually slightly different between the ranked and rankless cases
	
	-- look for an event table entry for the ranked version
	if rankedkey then
		local de = self:CopyTable( DetectedEvent )
		de.key = rankedkey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		
		if SpeakinSpellSavedData.ShowSpellEvents then
			if de.rank ~= "" then
				self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink.." ("..de.rank..")") ) --TODO: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
			else
				self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink) ) --TODO: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
			end
		end

		if de.EventTableEntry then
			SpeakForRankless = false
			if self:AllowSpeakForSpell( de ) then
				self:SpeakForSpell( de )
			end
		else
			-- record a new event for this ranked key
			self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent( de )
		end
	end
	
	-- check for a rankless version, but don't speak for both
	if ranklesskey then
		local de = self:CopyTable( DetectedEvent )
		de.key = ranklesskey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		
		if SpeakinSpellSavedData.ShowSpellEvents then
			self:Print( self:ProcessSubstitutionVars(L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink.." (Any Rank)") ) --TODO: add <target> and consider <displaylink> to encapsulate "L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.SpellLink"
		end
		
		if SpeakinSpellSavedData.EventTable[ de.key ] then
			if SpeakForRankless then
				if self:AllowSpeakForSpell( de ) then
					self:SpeakForSpell( de )
				end
			end
		else
			-- record a new RANKLESS event for this rankless key
			de.rank = "" -- to make the new event rankless -- NOTE: don't erase this before speaking, to maintain validity of <spellrank> variable
			de.DisplayName = self:FormatDisplayName( de ) -- update display name to show "any rank"
			self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent(de)
		end
	end

end


-------------------------------------------------------------------------------
-- SLASH COMMAND HANDLERS
-------------------------------------------------------------------------------


function SpeakinSpell:RegisterChatCommands()
	-- NOTE: support both localized and default English slash commands
	local SlashCommands = {
		L["/SpeakinSpell"],
		  "/SpeakinSpell",
		L["/speakinspell"],
		  "/speakinspell",
		L["/SS"],
		  "/SS",
		L["/ss"],
		  "/ss",
	}
	
	for k,v in pairs(SlashCommands) do
		self:RegisterChatCommand(v,"OnSlashCommand")
	end
end


function SpeakinSpell:OnSlashCommand(input)
	local funcname = "OnSlashCommand"

	input = string.lower(input)
	self:DebugMsg(funcname,"OnSlashCommand:"..input)
	
	-- if nothing is specified, then choose a default command
	if input == nil or input == "" then
		input = "options"
	else
		-- check for commands that require special parsing
		
		-- check for /ss ad /w stonaria
		local whisperad = self:StartsWith(input, "ad /w ")
		local len = string.len("ad /w ")
		if not whisperad then -- Localized check
			whisperad = self:StartsWith(input, L["/advertise"].." /w ")
			len = string.len(L["/advertise"].." /w ")
		end
		if whisperad then
			local target = string.sub(input, len+1, -1) --from len to the end of the string
			self:Advertise("WHISPER", target)
			return
		end
		
		-- check for /ss macro something
		local macro = self:StartsWith(input, "macro")
		if not macro then -- Localized check
			macro = self:StartsWith(input, L["/macro"])
		end
		if macro then
			self:DebugMsg(funcname, "User Macro input:"..input)
			local DetectedEventStub = {
				name = input, -- use the entire lowercase input = "macro something"
				rank = "",
				type = "MACRO",
				caster = UnitName("player"),
				target = self:GetDefaultTarget(),
			}
			self:OnEventDetected( DetectedEventStub )
			return
		end
		
	end
	
	-- table of supporting slash commands
	-- NOTE: support both English and current locale
	--	non-English game clients can typically use both translated and untranslated slash commands
	--	for built-in functions like /bg chat, etc, so it makes sense here as well
	local SlashCommandTable = {
		-- /ss options	- also show options
		[L["/options"]] = function() self:ShowOptions() end,
		[   "options" ] = function() self:ShowOptions() end,
		
		-- /ss create	- also Create New... interface
		[L["/create"]] = function() self:ShowCreateNew() end,
		[   "create" ] = function() self:ShowCreateNew() end,
		
		-- /ss create	- also Create New... interface
		[L["/messages"]] = function() self:ShowMessageOptions() end,
		[   "messages" ] = function() self:ShowMessageOptions() end,
		
		-- /ss help 	- show the help
		[L["/help"]] = function() self:ShowHelp() end,
		[   "help" ] = function() self:ShowHelp() end,
		
		-- /ss reset	- reset to defaults
		[L["/reset"]] = function() self:ResetToDefaults(nil) end,
		[   "reset" ] = function() self:ResetToDefaults(nil) end,
		
		-- /ss advertise- show an advertisement for SpeakinSpell
		[L["/advertise"]] = function() self:Advertise("") end,
		[   "advertise" ] = function() self:Advertise("") end,
		
		[L["/advertise"].." /s"] = function() self:Advertise("SAY") end,
		[   "advertise" .." /s"] = function() self:Advertise("SAY") end,
		
		[L["/advertise"].." /p"] = function() self:Advertise("PARTY") end,
		[   "advertise" .." /p"] = function() self:Advertise("PARTY") end,
		
		[L["/advertise"].." /ra"] = function() self:Advertise("RAID") end,
		[   "advertise" .." /ra"] = function() self:Advertise("RAID") end,
	}
	
	-- what did the user enter?
	Func = SlashCommandTable[input]
	
	-- process the command
	if Func then
		Func(input)
	else -- in all other cases, show help
		-- go straight to the slash commands page
		-- NOTE: A Blizzard Bug won't open a tier 3 page unless the tree has already been expanded to show tier 2
		--		so we'll try to show the parent frame first
		-- NOTE: I tried to work aroubd it by first opening the root page, but it stayed there
		--		I guess there's a timing issue with the windows messages (?) that makes it stick on the first one
		--		or maybe you're not allowed and it stops executing when you call this function (?)
		--self:ShowHelp() -- whatever, it didn't work, see notes above
		-- the slash commands page is now tier 2 to avoid this issue and other general bugginess with tier 3 pages
		InterfaceOptionsFrame_OpenToCategory(HELPFILE.SLASH_COMMANDS)
	end
end


-------------------------------------------------------------------------------
-- RANDOM SPEECHES - THE CORE FUNCTIONALITY
-------------------------------------------------------------------------------

	
function SpeakinSpell:SelectRandomMessage(DetectedEvent)
	assert(DetectedEvent.EventTableEntry and DetectedEvent.EventTableEntry.Messages) -- this function should not be called unless we have a valid table for this spell
	
	-- select a random message from the user-specific table
	
	local sizelist = #(DetectedEvent.EventTableEntry.Messages)
	local tempnum = math.random(1, sizelist);
	
	msg = DetectedEvent.EventTableEntry.Messages[tempnum]
	assert(msg)
	assert(type(msg) == "string")
	
	-- avoid repeating the same message we used last time
	-- unless there's only one in the list
	while sizelist > 1 and msg == self.RuntimeData.LastMessage[DetectedEvent.key] do
		tempnum = math.random(1, sizelist);
		msg = DetectedEvent.EventTableEntry.Messages[tempnum]
	end
	
	-- remember we selected this, so we don't repeat it next time
	self.RuntimeData.LastMessage[DetectedEvent.key] = msg
	
	return msg
end


function SpeakinSpell:GetChatChannelForSpell(EventTableEntry)
	-- NOTE: the EventTableEntry.Channels table stores channel names that are compatible with SendChatMessage
	--	see also CurrentMessagesGUI_OnChannelSelect
	if (IsActiveBattlefieldArena()) then
		return EventTableEntry.Channels.Arena
	elseif (GetNumBattlefieldScores() > 0) then
		return EventTableEntry.Channels.BG
	elseif (GetNumRaidMembers() > 0) then
		return EventTableEntry.Channels.Raid
	elseif (GetNumPartyMembers() > 0) then
		return EventTableEntry.Channels.Party
	else -- solo
		return EventTableEntry.Channels.Solo
	end
end


function SpeakinSpell:DoSubstitution(var, DetectedEvent)
	
	-- NOTE: also support the english variable names if running in a non-english locale
	--		this is meant to help support the fact that documentation in-game is lacking for this variable list
	--		it's also standard for non-English clients to support English commands

	local Substitutions = {
		[                "player"   ] = UnitName("player"), -- the player's name / your name
		[string.lower(L["<player>"])] = UnitName("player"), -- the player's name / your name
		
		[                "caster"   ] = DetectedEvent.caster, -- the caster's name
		[string.lower(L["<caster>"])] = DetectedEvent.caster, -- the caster's name
		
		[                "target"   ] = DetectedEvent.target, -- the target's name
		[string.lower(L["<target>"])] = DetectedEvent.target, -- the target's name
		
		[                "focus"   ] = UnitName("focus"), -- the focus target's name
		[string.lower(L["<focus>"])] = UnitName("focus"), -- the focus target's name
		
		[                "targetclass"   ] = DetectedEvent.targetclass, -- the target's name
		[string.lower(L["<targetclass>"])] = DetectedEvent.targetclass, -- the target's name
		
		[                "targetrace"   ] = DetectedEvent.targetrace, -- the target's name
		[string.lower(L["<targetrace>"])] = DetectedEvent.targetrace, -- the target's name
		
		[                "spellname"   ] = DetectedEvent.name,
		[string.lower(L["<spellname>"])] = DetectedEvent.name,
		
		[                "spellrank"   ] = DetectedEvent.rank,
		[string.lower(L["<spellrank>"])] = DetectedEvent.rank,
		
		[                "spelllink"   ] = DetectedEvent.SpellLink,
		[string.lower(L["<spelllink>"])] = DetectedEvent.SpellLink,
		
		[                "pet"   ] = DetectedEvent.pet,
		[string.lower(L["<pet>"])] = DetectedEvent.pet,
		
		-- auto-substitutions like MS Word would use
		[                "TM"   ] = "™",	-- Trademark symbol
		[string.lower(L["<TM>"])] = "™",	-- Trademark symbol
		
		[                "R"   ] = "®",	-- Registered Trademark symbol
		[string.lower(L["<R>"])] = "®",	-- Registered Trademark symbol
		
		[                "C"   ] = "©",	-- Copyright (C)
		[string.lower(L["<C>"])] = "©",	-- Copyright (C)
	}

	-- do case insensitive lookup of variable names
	local varkey = string.lower(var)

	-- check for var = "<target|___>", or localized equivalent
	-- this should normally work the same as <target>
	-- however if the target is the player, then show ___ instead of the player's own name
	-- For example "Hey no hitting on <target|me>!" results in
	-- "Hey no hitting on Fred!" if cast on Fred
	-- "Hey no hitting on me!" if cast on yourself

	local ndxPipe = string.find(varkey, "|")
	if ndxPipe and (ndxPipe >= 0) then
		varkey = string.sub(varkey, 0, ndxPipe-1)
		local me = string.sub(var, ndxPipe+1) -- NOTE: use original Case Sensitive var name to get "me"
		-- behavior of string.sub seems inconsistent, sometimes ndxPipe+1 gets the | character, which then throws a LUA error
		-- but ndxPipe+2 sometimes misses the first letter of "me"
		-- so we use ndxPipe+1 and check for the leading |
		if string.find(me,"|") then
			me = string.sub(me, 2)
		end
		-- use "me" instead of my own name if targetting myself
		if (me ~= "") then
			-- <target|me>
			if ( (varkey == "target") or (varkey == L["<target>"]) ) and (DetectedEvent.target == UnitName("player")) then
				Substitutions[varkey] = me
			end
			-- <caster|I>
			if ( (varkey == "caster") or (varkey == L["<caster>"]) ) and (DetectedEvent.caster == UnitName("player")) then
				Substitutions[varkey] = me
			end
		end

	end

	-- perform the substitution
	if Substitutions[varkey] then
		return Substitutions[varkey]
	end

	-- if we got this far then we did not find a value for this variable
	-- so return the variable unchanged, with wrapping < > restored
	return "<"..var..">"
end


function SpeakinSpell:ProcessSubstitutionVars(msg, DetectedEvent)
	return string.gsub(msg, "<(.-)>", 
		function(var) 
			return self:DoSubstitution(var, DetectedEvent) 
		end
	)
end


function SpeakinSpell:SpeakForSpell(DetectedEvent)
	-- determine which channel we will speak in
	channel = self:GetChatChannelForSpell(DetectedEvent.EventTableEntry)
	
	--TODOFUTURE: make it an option to whisper yourself - but be wary of the auto self-cast logic
	local WhisperTarget = DetectedEvent.EventTableEntry.WhisperTarget and DetectedEvent.target and (DetectedEvent.target ~= "") and (DetectedEvent.target ~= "Unknown") and (DetectedEvent.target ~= UnitName("player")) and UnitIsFriend("player",DetectedEvent.target)

	if not channel and not WhisperTarget then
		-- this spell's speeches have no target for the current situation
		self:DebugMsg("SpeakForSpell","current scenario has no message targets for " .. DetectedEvent.SpellLink)
		return
	end
	
	msg = self:SelectRandomMessage(DetectedEvent)
	
	-- process substitution variables
	msg = self:ProcessSubstitutionVars(msg, DetectedEvent)

	-- record the time when we spoke for this spell
	self.RuntimeData.LastMessageTime[DetectedEvent.key] = GetTime()
	
	-- say the message to the chat, if applicable
	if channel then
		SendChatMessage(msg, channel);
	end
	
	-- send the message to a whisper to the target, if applicable
	if WhisperTarget then
		SendChatMessage(msg, "WHISPER", nil, DetectedEvent.target);
	end
end


